import abc


class File(abc.ABC):
    def __init__(self, content: str):
        self.content = content.strip()

    def header(self) -> list[str]:
        return []

    def records(self) -> list[str]:
        return []

    def footer(self) -> list[str]:
        return []


class CsvFile(File):
    def header(self):
        return self.content.split("\n")[:1]

    def records(self):
        return self.content.split("\n")[1:]

    def footer(self):
        return []


class TextFile(CsvFile):
    pass


class GeojsonFile(File):
    def header(self):
        lines = self.content.split("\n")
        return [lines[0].split("[", 1)[0] + "["]

    def footer(self):
        lines = self.content.split("\n")
        return ["]" + "]".join(lines[-1].split("]")[2:]).removesuffix(",")]

    def records(self):
        lines = self.content.split("\n")
        return [
            lines[0].split("[", 1)[1],
            *[l for l in lines[1:-1]],
            "]".join(lines[-1].split("]")[:2]).removesuffix(","),
        ]


class KmlFile(File):
    def header(self):
        return self.content.split("\n")[:13]

    def records(self):
        return self.content.split("\n")[13:-3]

    def footer(self):
        return self.content.split("\n")[-3:]


class XmlFile(File):
    def header(self):
        return self.content.split("\n")[:3]

    def records(self):
        return self.content.split("\n")[3:-2]

    def footer(self):
        return self.content.split("\n")[-2:]


class QuakemlFile(XmlFile):
    pass


FILE_FMTS = {
    "csv": CsvFile,
    "text": TextFile,
    "geojson": GeojsonFile,
    "kml": KmlFile,
    "xml": XmlFile,
    "quakeml": QuakemlFile,
    None: GeojsonFile,
}


def get_file(fmt: str, content: str):
    file_fmt = FILE_FMTS[fmt]
    return file_fmt(content)


def join_files(files: list[File]) -> File:
    file_fmt = type(files[0])
    content = concat_header_records_footer(
        files[0].header(),
        sum([f.records() for f in files], []),
        files[-1].footer(),
    )
    return file_fmt(content)


def concat_header_records_footer(header: list[str], records: list[str], footer: list[str]):
    return "\n".join(header + records + footer)



from datetime import datetime


def check_time_field_is_valid(time_value: str):
    try:
        _ = datetime.fromisoformat(time_value)
    except Exception as exc:
        raise ValueError(f"Invalid time given (received {time_value})") from exc



from copy import copy
from dataclasses import asdict, dataclass, field
from typing import Any

from quaker_db.file import FILE_FMTS
from quaker_db.utils import check_time_field_is_valid


@dataclass
class Query:
    format: str = field(default=None)
    # Time
    endtime: str = field(default=None)
    starttime: str = field(default=None)
    updatedafter: str = field(default=None)
    # Location (rectangle)
    minlatitude: float = field(default=None)
    maxlatitude: float = field(default=None)
    minlongitude: float = field(default=None)
    maxlongitude: float = field(default=None)
    # Location (circle)
    latitude: float = field(default=None)
    longitude: float = field(default=None)
    maxradius: float = field(default=None)
    maxradiuskm: float = field(default=None)
    # Other
    catalog: str = field(default=None)
    contributor: str = field(default=None)
    eventid: str = field(default=None)
    includeallmagnitudes: bool = field(default=None)
    includeallorigins: bool = field(default=None)
    includearrivals: bool = field(default=None)
    includedeleted: bool = field(default=None)
    includesuperseded: bool = field(default=None)
    limit: int = field(default=None)
    maxdepth: float = field(default=None)
    maxmagnitude: float = field(default=None)
    mindepth: float = field(default=None)
    minmagnitude: float = field(default=None)
    offset: int = field(default=None)
    orderby: str = field(default=None)
    # Extensions
    alertlevel: str = field(default=None)
    callback: str = field(default=None)
    eventtype: str = field(default=None)
    jsonerror: bool = field(default=None)
    kmlanimated: bool = field(default=None)
    kmlcolorby: str = field(default=None)
    maxcdi: float = field(default=None)
    maxgap: float = field(default=None)
    maxmmi: float = field(default=None)
    maxsig: int = field(default=None)
    mincdi: float = field(default=None)
    minfelt: int = field(default=None)
    mingap: float = field(default=None)
    minsig: int = field(default=None)
    producttype: str = field(default=None)
    productcode: str = field(default=None)
    reviewstatus: str = field(default=None)

    def __post_init__(self):
        self.validate()

    def validate(self):
        for time_field in ["starttime", "endtime", "updatedafter"]:
            if (time := getattr(self, time_field)) is not None:
                check_time_field_is_valid(time)
        if not any(
            v == self.orderby for v in [None, "time", "time-asc", "magnitude", "magnitude-asc"]
        ):
            raise ValueError(f"Invalid orderby value: '{self.orderby}'.")
        if not any(v == self.format for v in FILE_FMTS.keys()):
            raise ValueError(f"Invalid format value: '{self.format}'.")

    def copy(self):
        return copy(self)

    def dict(self, include_nones: bool = False) -> dict[str, Any]:
        query_dict = asdict(self)
        if include_nones:
            return query_dict
        return {k: v for k, v in query_dict.items() if v is not None}



from argparse import ArgumentParser

from quaker_db.client import Client

DOCS = {
    "Formats": {
        "format": (
            str,
            "(default: quakeml) Specify the output format, one of csv, geojson, kml, quakeml, csv, text, xml",
        ),
    },
    "Time": {
        "endtime": (
            str,
            "(default: present time) Limit to events on or before the specified end time. NOTE: All times use ISO8601 Date/Time format. Unless a timezone is specified, UTC is assumed",
        ),
        "starttime": (
            str,
            "(default: NOW - 30 days) Limit to events on or after the specified start time. NOTE: All times use ISO8601 Date/Time format. Unless a timezone is specified, UTC is assumed",
        ),
        "updatedafter": (
            str,
            "(default: null) Limit to events updated after the specified time.  NOTE: All times use ISO8601 Date/Time format. Unless a timezone is specified, UTC is assumed",
        ),
    },
    "Location (rectangle)": {
        "minlatitude": (
            float,
            "(default: -90) [-90,90] degrees, Limit to events with a latitude larger than the specified minimum. NOTE: min values must be less than max values",
        ),
        "minlongitude": (
            float,
            "(default: -180) [-360,360] degrees, Limit to events with a longitude larger than the specified minimum. NOTE: rectangles may cross the date line by using a minlongitude < -180 or maxlongitude > 180. NOTE: min values must be less than max values",
        ),
        "maxlatitude": (
            float,
            "(default: 90) [-90,90] degrees, Limit to events with a latitude smaller than the specified maximum. NOTE: min values must be less than max values",
        ),
        "maxlongitude": (
            float,
            "(default: 180) [-360,360] degrees, Limit to events with a longitude smaller than the specified maximum. NOTE: rectangles may cross the date line by using a minlongitude < -180 or maxlongitude > 180. NOTE: min values must be less than max values",
        ),
    },
    "Location (circle)": {
        "latitude": (
            float,
            "(default: null) [-90,90] degrees, Specify the latitude to be used for a radius search",
        ),
        "longitude": (
            float,
            "(default: null) [-180,180] degrees, Specify the longitude to be used for a radius search",
        ),
        "maxradius": (
            float,
            "(default: 180) [0, 180] degrees, Limit to events within the specified maximum number of degrees from the geographic point defined by the latitude and longitude parameters. NOTE: This option is mutually exclusive with maxradiuskm and specifying both will result in an error",
        ),
        "maxradiuskm": (
            float,
            "(default: 20001.6) [0, 20001.6] km, Limit to events within the specified maximum number of kilometers from the geographic point defined by the latitude and longitude parameters. NOTE: This option is mutually exclusive with maxradius and specifying both will result in an error",
        ),
    },
    "Other": {
        "catalog": (
            str,
            "(default: null) Limit to events from a specified catalog. Use the Catalogs Method to find available catalogs. NOTE: when catalog and contributor are omitted, the most preferred information from any catalog or contributor for the event is returned",
        ),
        "contributor": (
            str,
            "(default: null) Limit to events contributed by a specified contributor. Use the Contributors Method to find available contributors. NOTE: when catalog and contributor are omitted, the most preferred information from any catalog or contributor for the event is returned",
        ),
        "eventid": (
            str,
            "(default: null) Select a specific event by ID; event identifiers are data center specific. NOTE: Selecting a specific event implies includeallorigins, includeallmagnitudes, and, additionally, associated moment tensor and focal-mechanisms are included",
        ),
        "includeallmagnitudes": (
            bool,
            "(default: false) Specify if all magnitudes for the event should be included, default is data center dependent but is suggested to be the preferred magnitude only. NOTE: because magnitudes and origins are strongly associated, this parameter is interchangeable with includeallmagnitude",
        ),
        "includeallorigins": (
            bool,
            "(default: false) Specify if all origins for the event should be included, default is data center dependent but is suggested to be the preferred origin only. NOTE: because magnitudes and origins are strongly associated, this parameter is interchangable with includeallmagnitude",
        ),
        "includedeleted": (
            bool,
            "(default: false) Specify if deleted products and events should be included. The value only returns only deleted events.  Deleted events otherwise return the HTTP status 409 Conflict.  NOTE: Only supported by the csv and geojson formats, which include status",
        ),
        "includesuperseded": (
            bool,
            "(default: false) Specify if superseded products should be included. This also includes all deleted products, and is mutually exclusive to the includedeleted parameter. NOTE: Only works when specifying eventid parameter",
        ),
        "maxmagnitude": (
            float,
            "(default: null) Limit to events with a magnitude smaller than the specified maximum",
        ),
        "mindepth": (
            float,
            "(default: -100) [-100, 1000] km Limit to events with depth more than the specified minimum",
        ),
        "minmagnitude": (
            float,
            "(default: null) Limit to events with a magnitude larger than the specified minimum",
        ),
        "orderby": (
            str,
            "Order the results. The allowed values are: time, time-asc, magnitude, magnitude-asc",
        ),
    },
}


def parse_args():
    parser = ArgumentParser()
    parser.add_argument(
        "argfile", nargs="?", help="Optional comma separated file with API arguments"
    )

    for group_name, params in DOCS.items():
        group = parser.add_argument_group(group_name + " arguments")

        for field_name, (field_type, doc) in params.items():
            add_arg_kwargs = dict(
                help=doc,
                required=False,
                default=None,
            )
            if field_type is bool:
                add_arg_kwargs["action"] = "store_true"
            else:
                add_arg_kwargs["type"] = field_type

            group.add_argument(
                "--" + field_name,
                **add_arg_kwargs,
            )

    return parser.parse_args()


def main():
    args = vars(parse_args())

    if (args_file := args.pop("args_file", None)) is not None:
        args = {}
        docs_types = {k: t for n, v in DOCS.items() for k, (t, _) in v.items()}
        with open(args_file, "r") as f:
            for line in f.readlines():
                k, v = line.split(":", 1)
                k, v = k.strip(), v.strip()
                t = docs_types[k]
                v = t(v)
                args[k] = v

    client = Client()
    result = client.execute(**args)

    print(result)


if __name__ == "__main__":
    main()



import logging
import random
from time import sleep

from requests.sessions import Session

from quaker_db.file import get_file, join_files
from quaker_db.globals import (
    BASE_URL,
    MAX_ATTEMPTS,
    NUM_WORKERS,
    RESPONSE_BAD_REQUEST,
    RESPONSE_NOT_FOUND,
)
from quaker_db.query import Query


class Client:
    def __init__(self):
        self.num_workers = NUM_WORKERS

        self.session = Session()
        self.logger = logging.getLogger(__name__)

    def execute(self, **kwargs):
        query = Query(**kwargs)
        return self._execute_sq(query)

    def _execute_sq(self, query: Query) -> str:
        query.limit = 20000
        query.offset = 1

        pages = []
        fetch_next_page = True
        while fetch_next_page:
            page = get_file(query.format, self._execute(query))
            pages.append(page)
            if len(page.records()) <= 20000:
                fetch_next_page = False
            else:
                query.offset += 20000

        return join_files(pages).content

    def _execute(self, query: Query) -> str:
        with self.session as session:
            for idx in range(MAX_ATTEMPTS):
                sleep(random.expovariate(1 + idx * 0.5))
                response = session.get(BASE_URL, params=query.dict())

                if response.status_code != RESPONSE_NOT_FOUND:
                    self._check_download_error(response)
                    return response.text.strip()

                self.logger.warning(f"No connection could be made, retrying ({idx}).")

        raise ConnectionAbortedError("Connection could not be established")

    def _check_download_error(self, response):
        if response.ok:
            return

        status = response.status_code
        msg = f"Unexpected response code on query ({status})."
        if status == RESPONSE_BAD_REQUEST:
            msg = f"Invalid query ({RESPONSE_BAD_REQUEST})."

        self.logger.error(msg)
        raise RuntimeError(msg)
